// When I started out with these LED displays I could not really program at all, but I managed to figure out a lot of this thanks to all the resources and tips available on the web. 
// This sketch takes a lot of inspiration and code snippets from the FastLED "NoisePlusPalette" example but significantly adapts it all.
// I at one point started with Thomasj152's code at https://www.instructables.com/Ping-Pong-Ball-LED-Clock/ . That code was  adapted significantly first for a 23x7 display and now this one (24x12). 
// It's by now probably unrecognisably different from the original, and I believe it draws on one of the DS3232RTC examples, although I cant find where it all came from
// It should run on an arbitrarily sized LED matrix although I never tried any different layouts. You'll also very rapidly run out of memory at larger sizes.
// Also draws on many code snippets found across the internet and help on forums (especially on how to use PROGMEM, without which all this would never fit in memory!!). Thanks for all the anonymous help out there!

#include <FastLED.h>
#include <DS3232RTC.h>
#include <Button.h>
#define NUM_LEDS 288
#define DATA_PIN 6
#define BRIGHTNESS 20
CRGB leds[NUM_LEDS];
#define columns 24
#define rows 12
DS3232RTC myRTC(false);
tmElements_t tm;

//variables used to calculate the LED positions to light up
uint8_t col;
uint8_t row;
int r;
uint8_t number = 0;
uint8_t xpos = 1;  //x=column
uint8_t ypos = 1;  //y=row

//Background variables
uint8_t backgroundcounter = 0;
#define stripethickness 20
static uint16_t noisex;
static uint16_t noisey;
static uint16_t noisez;
uint16_t speed = 5;
uint16_t scale = 50;
//Colour palette array
CRGBPalette16 currentPalette(ForestColors_p);
//Array to hold Perlin noise for the background
uint8_t noise[columns][rows];
uint8_t BrightnessDimming = 75;  //Adjusts background brightness for legibility; 0= same as global brightness, 255 = black

// Variables used to display the time
uint8_t hr10;
uint8_t hr1;
uint8_t min10;
uint8_t min1;

//Buttons
Button cyclefont(11);
uint8_t font = 2;
Button cyclebackground(9);
uint8_t background = 0;
Button nextpalette(7);
uint8_t palette = 0;

//Custom Palettes
const TProgmemPalette16 myRedWhiteBluePalette_p PROGMEM ={    CRGB::Red,    CRGB::Gray,    CRGB::Blue,    CRGB::Black,    CRGB::Red,    CRGB::Gray,    CRGB::Blue,    CRGB::Black,    CRGB::Red,    CRGB::Red,    CRGB::Gray,    CRGB::Gray,    CRGB::Blue,    CRGB::Blue,    CRGB::Black,    CRGB::Black};
const TProgmemPalette16 PurpleAndGreenPalette_P PROGMEM ={   CRGB::Green,  CRGB::Green,  CRGB::Black,  CRGB::Black,  CRGB::Purple,  CRGB::Purple,  CRGB::Black,  CRGB::Black,  CRGB::Green,  CRGB::Green,  CRGB::Black,  CRGB::Black,  CRGB::Yellow,  CRGB::Purple,  CRGB::Black,  CRGB::Black,};
const TProgmemPalette16 BlackWhitePalette_p PROGMEM = {  CRGB::Gray,  CRGB::Black,  CRGB::Black,  CRGB::Black,  CRGB::Black,  CRGB::Gray,  CRGB::Black,  CRGB::Black,  CRGB::Gray,  CRGB::Gray,  CRGB::Black,  CRGB::Black,  CRGB::Black,  CRGB::Black,  CRGB::Black,};
const TProgmemPalette16 RedGreenBlue_p PROGMEM = {  CRGB::Red,  CRGB::Red,  CRGB::Red,  CRGB::Red,  CRGB::Green,  CRGB::Green,  CRGB::Green,  CRGB::Green,  CRGB::Blue,  CRGB::Blue,  CRGB::Blue,  CRGB::Blue,};

// Fonts
const PROGMEM uint8_t digits[60][48] = {
  { 14, 26, 38, 50, 3, 15, 51, 63, 4, 40, 64, 5, 29, 41, 65, 6, 42, 66, 7, 31, 67, 8, 32, 44, 68, 9, 33, 69, 10, 70, 23, 35, 47, 59 },  //font 0 = Phat_outline
  { 26, 38, 50, 15, 51, 4, 52, 5, 17, 29, 53, 18, 54, 19, 55, 20, 56, 9, 21, 57, 69, 10, 70, 11, 23, 35, 47, 59, 71 },
  { 14, 26, 38, 50, 3, 63, 4, 16, 28, 64, 29, 65, 18, 54, 7, 43, 8, 32, 44, 56, 68, 9, 69, 10, 70, 11, 23, 35, 47, 59 },
  { 14, 26, 38, 50, 3, 63, 4, 16, 28, 64, 29, 65, 18, 54, 19, 55, 8, 20, 32, 68, 9, 69, 10, 70, 23, 35, 47, 59 },
  { 26, 38, 50, 27, 51, 16, 52, 17, 53, 6, 42, 66, 7, 67, 8, 20, 32, 44, 68, 45, 69, 46, 70, 47, 59, 71 },
  { 2, 14, 26, 38, 50, 62, 3, 63, 4, 64, 5, 29, 41, 53, 6, 42, 19, 55, 32, 68, 9, 21, 33, 69, 10, 70, 23, 35, 47, 59 },
  { 26, 38, 50, 62, 15, 63, 16, 52, 64, 5, 41, 6, 42, 54, 7, 67, 8, 32, 68, 9, 33, 45, 69, 10, 70, 23, 35, 47, 59 },
  { 2, 14, 26, 38, 50, 62, 3, 63, 4, 16, 28, 40, 64, 41, 65, 42, 66, 31, 67, 32, 56, 21, 57, 22, 46, 23, 35, 47 },
  { 14, 26, 38, 50, 3, 63, 4, 28, 40, 64, 5, 41, 65, 18, 54, 7, 67, 8, 32, 44, 68, 9, 45, 69, 10, 70, 23, 35, 47, 59 },
  { 14, 26, 38, 50, 3, 63, 4, 28, 40, 64, 5, 41, 65, 6, 66, 19, 31, 67, 32, 68, 9, 21, 57, 10, 58, 11, 23, 35, 47 },                                                                    //end font 0
  { 14, 26, 38, 50, 3, 15, 27, 39, 51, 63, 4, 16, 28, 52, 64, 5, 17, 53, 65, 6, 18, 54, 66, 7, 19, 55, 67, 8, 20, 56, 68, 9, 21, 45, 57, 69, 10, 22, 34, 46, 58, 70, 23, 35, 47, 59 },  //font 1 = Phat_filled
  { 38, 50, 27, 39, 51, 16, 28, 40, 52, 17, 29, 41, 53, 30, 42, 54, 31, 43, 55, 32, 44, 56, 33, 45, 57, 34, 46, 58, 35, 47, 59 },
  { 14, 26, 38, 50, 3, 15, 27, 39, 51, 63, 4, 16, 52, 64, 41, 53, 65, 30, 42, 54, 19, 31, 43, 8, 20, 32, 9, 21, 10, 22, 34, 46, 58, 70, 11, 23, 35, 47, 59, 71 },
  { 14, 26, 38, 50, 3, 15, 27, 39, 51, 63, 4, 16, 52, 64, 41, 53, 65, 30, 42, 54, 43, 55, 8, 20, 44, 56, 68, 9, 21, 33, 45, 57, 69, 10, 22, 34, 46, 58, 70, 23, 35, 47, 59 },
  { 38, 50, 39, 51, 4, 16, 40, 52, 5, 17, 41, 53, 6, 18, 42, 54, 7, 19, 31, 43, 55, 67, 8, 20, 32, 44, 56, 68, 45, 57, 46, 58, 47, 59 },
  { 2, 14, 26, 38, 50, 62, 3, 15, 27, 39, 51, 63, 4, 16, 5, 17, 29, 41, 53, 18, 30, 42, 54, 66, 55, 67, 56, 68, 9, 21, 57, 69, 10, 22, 34, 46, 58, 70, 23, 35, 47, 59 },
  { 26, 38, 50, 15, 27, 39, 51, 16, 28, 40, 5, 17, 29, 6, 18, 30, 42, 54, 7, 19, 43, 55, 67, 8, 20, 56, 68, 9, 21, 57, 69, 10, 22, 34, 46, 58, 70, 23, 35, 47, 59 },
  { 2, 14, 26, 38, 50, 62, 3, 15, 27, 39, 51, 63, 52, 64, 41, 53, 65, 42, 54, 43, 55, 32, 44, 56, 33, 45, 34, 46, 35, 47 },
  { 14, 26, 38, 50, 3, 15, 27, 39, 51, 63, 4, 16, 52, 64, 5, 17, 29, 53, 65, 18, 30, 42, 54, 7, 19, 31, 43, 55, 67, 8, 20, 56, 68, 9, 21, 33, 57, 69, 10, 22, 34, 46, 58, 70, 23, 35, 47, 59 },
  { 14, 26, 38, 50, 3, 15, 27, 39, 51, 63, 4, 16, 52, 64, 5, 17, 53, 65, 6, 18, 30, 54, 66, 19, 31, 43, 55, 67, 44, 56, 68, 33, 45, 57, 22, 34, 46, 58, 23, 35, 47 },  //end font 1
  { 16, 28, 40, 52, 5, 65, 6, 42, 66, 7, 31, 43, 67, 8, 68, 21, 33, 45, 57 },                                                                                          //font 2 = small_phat_outline
  { 40, 52, 29, 41, 53, 18, 30, 54, 31, 55, 32, 56, 33, 45, 57 },
  { 16, 28, 40, 52, 5, 65, 6, 18, 66, 19, 43, 55, 8, 68, 9, 21, 33, 45, 57, 69 },
  { 16, 28, 40, 52, 5, 65, 18, 54, 19, 31, 67, 8, 68, 21, 33, 45, 57 },
  { 40, 52, 29, 53, 18, 54, 66, 7, 67, 8, 20, 32, 44, 68, 45, 57, 69 },
  { 4, 16, 28, 40, 52, 64, 5, 65, 6, 42, 54, 19, 67, 8, 68, 21, 33, 45, 57 },
  { 16, 28, 40, 52, 5, 65, 6, 30, 42, 54, 7, 67, 8, 68, 21, 33, 45, 57 },
  { 4, 16, 28, 40, 52, 64, 5, 65, 6, 18, 30, 66, 31, 55, 20, 44, 21, 33, 45 },
  { 16, 28, 40, 52, 5, 65, 18, 42, 54, 7, 67, 8, 68, 21, 33, 45, 57 },
  { 16, 28, 40, 52, 5, 65, 6, 66, 19, 31, 43, 67, 8, 68, 21, 33, 45, 57 },  //end font 2
  { 0 }, // font 3: count the lines
  { 29, 30, 31, 32, 33 },
  { 29, 30, 7, 19, 31, 43, 55, 32, 33 },
  { 29, 53, 30, 42, 7, 19, 31, 43, 55, 20, 32, 9, 33 },
  { 5, 29, 53, 18, 30, 42, 7, 19, 31, 43, 55, 20, 32, 44, 9, 33, 57 },
  { 5, 17, 29, 41, 53, 6, 54, 7, 55, 8, 56, 9, 21, 33, 45, 57 },
  { 5, 17, 29, 41, 53, 6, 30, 54, 7, 31, 55, 8, 32, 56, 9, 21, 33, 45, 57 },
  { 5, 17, 29, 41, 53, 6, 30, 54, 7, 19, 31, 43, 55, 8, 32, 56, 9, 21, 33, 45, 57 },
  { 5, 17, 29, 41, 53, 6, 30, 42, 54, 7, 19, 31, 43, 55, 8, 20, 32, 56, 9, 21, 33, 45, 57 },
  { 5, 17, 29, 41, 53, 6, 18, 30, 42, 54, 7, 19, 31, 43, 55, 8, 20, 32, 44, 56, 9, 21, 33, 45, 57 },  //end font 3
  { 16, 28, 40, 52, 5, 65, 6, 66, 7, 67, 8, 68, 21, 33, 45, 57 },                                     // font 4 = small_lines
  { 40, 41, 42, 43, 44, 45 },
  { 16, 28, 40, 52, 5, 65, 42, 54, 19, 31, 8, 9, 21, 33, 45, 57, 69 },
  { 16, 28, 40, 52, 5, 65, 30, 42, 54, 67, 8, 68, 21, 33, 45, 57 },
  { 40, 29, 41, 18, 42, 7, 43, 8, 20, 32, 44, 56, 68, 45 },
  { 4, 16, 28, 40, 52, 64, 5, 18, 30, 43, 55, 8, 68, 21, 33, 45, 57 },
  { 16, 28, 40, 52, 5, 65, 6, 7, 19, 31, 43, 55, 67, 8, 68, 21, 33, 45, 57 },
  { 4, 16, 28, 40, 52, 64, 65, 54, 43, 55, 44, 45 },
  { 16, 28, 40, 52, 5, 65, 18, 30, 42, 54, 7, 67, 8, 68, 21, 33, 45, 57 },
  { 16, 28, 40, 52, 5, 65, 6, 18, 30, 42, 54, 66, 67, 8, 68, 21, 33, 45, 57 },  //end font 4
  { 0 }, //font 5: blank font
  { 0 },
  { 0 },
  { 0 },
  { 0 },
  { 0 },
  { 0 },
  { 0 },
  { 0 },
  { 0 },  //font 5 = blank
};

void setup() {
  Serial.begin(9600);
  FastLED.addLeds<WS2812B, DATA_PIN, GRB>(leds, NUM_LEDS);
  FastLED.setBrightness(BRIGHTNESS);
  FastLED.clear();
  delay(500);
  cyclefont.begin();
  cyclebackground.begin();
  nextpalette.begin();
  noisex = random16();
  noisey = random16();
  noisez = random16();
}

void loop() {
  RTC.read(tm);
  if (background == 1) {    fillnoise8();    } //fill background with Perlin noise
  if (background == 2) {    PaletteBands();  } //diagonal bands, palettised
  if (background == 3) {    Bands();         } //red green and blue bands
  fadeToBlackBy(leds, NUM_LEDS, BrightnessDimming);  //Dim the background so the numbers are more easily visible
  displaytime();
  FastLED.show();
  FastLED.clear();
 
  if (cyclefont.pressed()) {
    font++;
    if (font > 5) font = 0;
 }
  if (cyclebackground.pressed()) {
    background++;
    if (background > 3) background = 0;
 }
  if (nextpalette.pressed()) {
    palette++;
    if (palette > 10) palette =0;
    if (palette ==0) {      currentPalette = OceanColors_p;      speed = 10;      scale = 50;    };
    if (palette ==1) {      currentPalette = PurpleAndGreenPalette_P;speed = 10;  scale = 50;    };
    if (palette ==2) {      currentPalette = ForestColors_p;     speed = 8;       scale = 100;   };
    if (palette ==3) {      currentPalette = RainbowColors_p;    speed = 4;       scale = 90;    };
    if (palette ==4) {      currentPalette = LavaColors_p;       speed = 8;       scale = 50;    };
    if (palette ==5) {      currentPalette = CloudColors_p;      speed = 4;       scale = 30;    };
    if (palette ==6) {      currentPalette = HeatColors_p;       speed = 10;      scale = 20;    };
    if (palette ==7) {      currentPalette = RainbowStripeColors_p; speed = 2;   scale = 10;    };
    if (palette ==8) {      currentPalette = myRedWhiteBluePalette_p; speed = 4; scale = 5;    };
    if (palette ==9) {      currentPalette = BlackWhitePalette_p; speed = 1; scale = 5;    };
    if (palette ==10){      currentPalette = RedGreenBlue_p;     speed = 10; scale = 50;    };
  }
  backgroundcounter++;
}

void displaytime() {
  hr10 = (tm.Hour / 10) + font * 10;
  hr1 = (tm.Hour % 10) + font * 10;
  min10 = (tm.Minute / 10) + font * 10;
  min1 = (tm.Minute % 10) + font * 10;
  Displaynumber(hr10, 1, 0);
  Displaynumber(hr1, 7, 2);
  Displaynumber(min10, 13, 0);
  Displaynumber(min1, 19, 2);
  if (xpos >= columns) xpos = 0;
  if (ypos > rows) ypos = 0;
}

void Displaynumber(uint8_t number, uint8_t xpos, uint8_t ypos) {
  for (uint8_t i = 0; i < 48; i++) {
    uint8_t pixelnumber = pgm_read_byte(&(digits[number][i]));
    if (pixelnumber == 0) return 32767;
    uint8_t x = (pixelnumber / 12) + xpos;
    uint8_t y = ((pixelnumber - 1) % 12) + ypos;
    if (x > columns) x = x - columns;
    if (y > rows) y = y - rows;
    r = { calcLEDposition(x, y) };
    leds[r] = CRGB::White;
    // leds[r] =- leds[r]; // This inverts colours instead of going for a single letter colour
  }
}

int calcLEDposition(uint8_t col, uint8_t row) {  //Calculate the position of the LED to light up. column & row 0 are not displayed, 'overscan'areas! This is used to hide away artefacts at position 0,0 which I got otherwise
  col = columns+1-col;
  if (row == 0 || col == 0) {    return 32767;  }
  if (col % 2 == 0) {    int r = (300 - ((col * 12) + row));    return r;  } 
  else {    int r = (300 - ((col * 12) + (13 - row)));    return r;  }
}

void Bands() { //old code from when I first started out, cant really remember te reasoning behind all this but it does seem to work ;)
  delay(50);
  for (uint8_t i = 1; i <= columns; i++) {
    for (uint8_t j = 1; j <= rows; j++) {
      uint8_t modulo = (i + j + backgroundcounter) % stripethickness;
      modulo = map(modulo, 0, stripethickness, 0, 3);
      modulo = map(modulo, 0, 3, 0, 255);
      int r{ calcLEDposition(i, j) };
      leds[r] = CHSV(modulo, 255, 255);
    }
  }
}

void PaletteBands() {
  delay(50);
  for (uint8_t i = 1; i <= columns; i++) {
    for (uint8_t j = 1; j <= rows; j++) {
      uint8_t value = (i + j + backgroundcounter) % stripethickness;
      value = map(value, 0, stripethickness, 0, 16);
      value = map(value, 0, 16, 0, 255);
      int r{ calcLEDposition(i, j) };
      CRGB colour = ColorFromPalette(currentPalette, value, value);
      leds[r] = colour;
      // leds[r] = CHSV(modulo, 255, 255);
    }
  }
}

void fillnoise8() {
  for (uint8_t i = 1; i <= columns; i++) {
    uint16_t ioffset = scale * i;
    for (uint8_t j = 1; j <= rows; j++) {
      uint16_t joffset = scale * j;
      uint8_t value = inoise8(noisex + ioffset, noisey + joffset, noisez);
      value = qsub8(value, 16);
      value = qadd8(value, scale8(value, 39));
      r = { calcLEDposition(i, j) };
      CRGB colour = ColorFromPalette(currentPalette, value, value);
      leds[r] = colour;
    }
  }
  noisez += speed;
  noisex += speed / 8;
  noisey += speed / 16;
}
